#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include "Block.hh"
#include "Header.hh"
#include "PSRFits.hh"
#include "Utils.hh"

using namespace Header;
using namespace Block;
using namespace FITS;
using std::string;

int main(int argc, char * argv[])
{
  std::string fname;
  bool usingHighLatitude = false;
  if(argc==2){
    fname = argv[1];
    if(fname=="-h") {
      std::cerr << "Usage: \"pvoconvert.x filename\"\n"
		<< "   or: \"pvoconvert.x -ph filename\" to set PH survey.\n"
		<< "   or: \"pvoconvert.x -h \" to see this message.\n";
      return 0;
    }
  }
  else if(argc==3){
    std::string opt = argv[1];
    fname = argv[2];
    if(opt == "-ph") {
      usingHighLatitude = true;
    }
    else {
      std::cerr << "Usage: \"pvoconvert.x filename\"\n"
		<< "   or: \"pvoconvert.x -ph filename\" to set PH survey.\n"
		<< "   or: \"pvoconvert.x -h \" to see this message.\n";
      return 1;
    }
  }
  else{
    std::cerr << "Usage: \"pvoconvert.x filename\"...\n"
	      << "   or: \"pvoconvert.x -ph filename\" to set PH survey.\n"
	      << "   or: \"pvoconvert.x -h \" to see this message.\n";
    return 1;
  }

  SURVEY currentSurvey = idSurvey(fname);
  if(currentSurvey == NONEXISTENT) {
    std::cerr << "File " << fname <<" does not exist.\n";
    std::cerr << "Exiting...\n";
    return 1;
  }
  
  if(usingHighLatitude){
    if(currentSurvey == PM) currentSurvey = PH;
    else{
      std::cerr << "Error: you selected the -h option, which means High Latitude survey.\n"
		<< "But " << fname << " is a " << surveyNames[currentSurvey]
		<<" file.\nExiting...\n";
      return 1;
    }
  }
  std::cerr << "File " << fname <<" : " << surveyNames[currentSurvey]<<"\n";
  if(currentSurvey == DUD){
    std::cerr << "Exiting...\n";
    return 1;
  }

  std::ifstream fin;
  fin.open(fname.c_str(), std::ios::binary);

  bool eof;
  DataBlock block(currentSurvey);
  block.readBlock(fin);
  
  int newBeam = block.getKeyword<int>(IBEAM);
  int oldBeam;
  bool newMoveFlag= block.telescopeOK();
  bool oldMoveFlag;

  PSRFits fitsfile;

  if(fitsfile.initialise(block) == FAILURE){
    std::cerr << "Could not open PSRFITS file "
	      << fitsfile.getFITSname()
	      << ".\nExiting...\n";
    return 1;
  }

  fitsfile.writeHistoryTapeFile(fname);
  fitsfile.writeHistoryCommand(argc,argv);
  
  do{

    if(fitsfile.getFilestate()==OPEN) fitsfile.writeSubintRow(block);

    oldBeam = newBeam;
    oldMoveFlag = newMoveFlag;
 	
    // We've done all we can for this block. Read a new one.
    eof = !block.readBlock(fin);

    if(!eof){ // if the file is still open
      
      bool fileFull = fitsfile.fitsFull(block) ;

      // get info on the new block
      newBeam     = block.getKeyword<int>(IBEAM);
      newMoveFlag = block.telescopeOK();
      
      if((newBeam != oldBeam) || fileFull ){
	//if we've moved, or it's a new beam, close the file.
	fitsfile.closeFile();
      }

      if(newMoveFlag){

	if(!oldMoveFlag){
	  // if this is true, then we were moving but now are not.
	  // this means it is a new file.
	  fitsfile.setNewFile();
	}

	if( fitsfile.getFilestate()==CLOSED ){
	  // if there is no file open, but we are not moving, 
	  //  open a new file
	  fitsfile.initialise(block);
	  
	  if(fitsfile.isNewFile()) {
	    fitsfile.writeHistoryTapeFile(fname);
	    fitsfile.writeHistoryCommand(argc,argv);
	  }

	}

      }

    }
	
  }while(!eof);
  
  fitsfile.closeFile();
  fin.close();

  return 0;

}
